# -*- coding: iso-8859-1 -*-
# Copyright (c) 2003-2006 ETH Zurich, D-MATL. Written by Lorenz Textor. All rights reserved.
# See also LICENSE.txt

import os

from urllib import urlencode
import marshal
import base64
from array import array

from MySQLdb import escape_string

from interfaces import IDBEditor

#Zope
from AccessControl import ClassSecurityInfo, Permissions, getSecurityManager
from Globals import InitializeClass
from Globals import DTMLFile
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
#from zope.app.file.image import Image
from App.ImageFile import ImageFile
from Products.Silva.helpers import add_and_edit
from Products.ZSQLMethods.SQL import SQL
from Products.ZSQLMethods.SQL import SQLConnectionIDs
from Products.PythonScripts.standard import newline_to_br, html_quote

from Acquisition import Implicit
from Globals import Persistent
from AccessControl.Role import RoleManager
from OFS.SimpleItem import Item

class DBEditor(Implicit, Persistent, RoleManager, Item):
        """
        DBEditor product class, implements IDBEditor interface.

        Contains all methods for the DBEditor.
        """

        __implements__ = IDBEditor
        
        meta_type='DBEditor'

        security=ClassSecurityInfo()

        manage_options = (
                {'label' : 'Edit', 'action' : 'manage_main'},
                {'label' : 'View', 'action' : ''},
                ) + RoleManager.manage_options + Item.manage_options

        manage_main = edit_tab = PageTemplateFile('www/DBEditorEditTab', globals(),
                                                  __name__='edit_tab')

## defaults for new attributes
	_error = ''
	_logTable = ''
	_showTableSelector = True
	_preselectTable = None

## init method
	def __init__(self, id, title, dbConnectionId):
		self.id = id
		self.title = title
		self._dbConnectionId = dbConnectionId

		self._error = ''
		self._logTable = ''
		self._showTableSelector = True
		self._preselectTable = None

		self._columnSettings = {}
		self._roles = {'viewSecretColumns':'Viewer +', 'editData':'Author', 'addRemoveData':'Editor'}
		self._nullString = 'NULL'
		self._footer = ''

## manage methods
	# methods starting with manage_ require always management rights therefore no security declaration necessary!
	def manage_edit(self):
		"""
		Saves changes made in the edit tab.
		"""
		req = self.REQUEST

		if ( getattr(req, 'dbConnectionSave', None ) ) :
		# save dbConnectionId
			# reset session
			self.saveDb('')
			dbConnectionId = req.dbConnectionId
			connect = self.setDbConnectionId(dbConnectionId)
			if not dbConnectionId :
				manage_tabs_message = 'ERROR: Please select a database connection.'
			else :
				if not connect :
					manage_tabs_message = 'ERROR: Could not find database connection "%s".' % dbConnectionId
				else :
					manage_tabs_message = 'Database connection changed to "%s".' % dbConnectionId
			return self.edit_tab(manage_tabs_message=manage_tabs_message)

		if ( getattr(req, 'generalSettingsSave', None ) ) :
		# save general settings
			manage_tabs_message = 'General settings saved.'
			if getattr(req,'showTableSelector',None) :
				self._showTableSelector = True
			else :
				self._showTableSelector = False
			if getattr(req,'preselectTable',None) :
				database = req.preselectTable[1:-1].split('`.`')[0]
				table = req.preselectTable[1:-1].split('`.`')[1]
				self._preselectTable = (database,table)
			else :
				self._preselectTable = None
			self._nullString = req.nullString
			self._footer = req.footer
			self._logTable = req.logTable
			if self._logTable and not self.checkLogTable(req.logTable) :
				self._logTable = ''
				manage_tabs_message += '<br />ERROR: The given table has a wrong structure for use with logging. Please see README for the right structure to be used.'
			return self.edit_tab(manage_tabs_message=manage_tabs_message)

		if ( getattr(req, 'accessSettingsSave', None ) ) :
		# save access settings
			if req.viewSecretColumns :
				self._roles['viewSecretColumns'] = req.viewSecretColumns
			else :
				if self._roles.has_key('viewSecretColumns') :
					del self._roles['viewSecretColumns']
			if req.editData :
				self._roles['editData'] = req.editData
			else :
				if self._roles.has_key('editData') :
					del self._roles['editData']
			if req.addRemoveData :
				self._roles['addRemoveData'] = req.addRemoveData
			else :
				if self._roles.has_key('addRemoveData') :
					del self._roles['addRemoveData']
			# tell the persistence machinery that a mutable attribute has changed
			self._p_changed = 1
			manage_tabs_message = 'Access settings saved.'
			return self.edit_tab(manage_tabs_message=manage_tabs_message)

		if ( getattr(req, 'tableSelectorChanged', None ) ) :
		# new db/table selected -> save it to session
			self.saveDb(req.selectedDatabase)
			self.saveTable(req.selectedTable)

		if ( getattr(req, 'columnSettingsSave', None ) ) :
		# save columnSettings
			columnOrder = req.columnOrder.split(',')
			columns = {}
			hiddenColumns = []
			secretColumns = []
			for value in req.form :
				# get mapping value
				if value.startswith('mapping_'):
					fieldname = value[8:]
					if not columns.has_key(fieldname):
						columns[fieldname] = {}
					columns[fieldname]['mapping'] = req.form[value]
				# get title value
				if value.startswith('title_'):
					fieldname = value[6:]
					if not columns.has_key(fieldname):
						columns[fieldname] = {}
					columns[fieldname]['title'] = req.form[value]
				# get help value
				if value.startswith('help_'):
					fieldname = value[5:]
					if not columns.has_key(fieldname):
						columns[fieldname] = {}
					columns[fieldname]['help'] = req.form[value]
				# get help value
                                if value.startswith('order_'):
					fieldname = value[6:]
					if not columns.has_key(fieldname):
						columns[fieldname] = {}
                                        value = req.form[value]
					columns[fieldname]['order'] = value and int(value) or 0
				# get hidden value -> we have to save it in an array and go through all columns afterwards
				#   because checkboxes which are not checked are not sent
				if value.startswith('hidden_'):
					if req.form[value] :
						hiddenColumns.append(value[7:])
				# get secret value
				if value.startswith('secret_'):
					if req.form[value] :
						secretColumns.append(value[7:])
			# save hidden and secret columns
			for column in columns :
				if column in hiddenColumns :
					columns[column]['hidden'] = 1
				else :
					columns[column]['hidden'] = 0
#				if column in secretColumns :
#					columns[column]['secret'] = 1
#				else :
#					columns[column]['secret'] = 0
			# save settings
			database = self.getSelectedDatabase()
			table = self.getSelectedTable()
			if not self._columnSettings.has_key(database) :
				# create new db entry
				self._columnSettings[database] = {}
			if not self._columnSettings[database].has_key(table) :
				# create new table entry if no table entry
				self._columnSettings[database][table] = {}
			elif not self._columnSettings[database][table].has_key('secretColumns') :
				# table entry in old format (<= v0.6(beta3)) -> save old settings for list and details view (secretColumns will be saved anyway afterwards)
				oldSettings = self._columnSettings[database][table]
				self._columnSettings[database][table] = {'list': oldSettings, 'details': oldSettings}
			if req.view == 'details' :
				self._columnSettings[database][table]['details'] = columns
			else :
				self._columnSettings[database][table]['list'] = columns
			self._columnSettings[database][table]['secretColumns'] = secretColumns
			# tell the persistence machinery that a mutable attribute has changed
			self._p_changed = 1
			# reset session (in case user switches to DBEditor)
			if req.SESSION.has_key('selectedColumns') :
				del req.SESSION['selectedColumns']
			if req.SESSION.has_key('fields') :
				del req.SESSION['fields']
			manage_tabs_message = 'Column settings saved.'
			return self.edit_tab(manage_tabs_message=manage_tabs_message)

		return self.edit_tab()

## PageTemplate methods
	security.declareProtected(Permissions.access_contents_information, 'mainTemplate')
	mainTemplate = PageTemplateFile('www/mainTemplate', globals())

	security.declareProtected(Permissions.access_contents_information, 'index_html')
	index_html = PageTemplateFile('www/index', globals(), __name__='index_html')

	security.declareProtected(Permissions.access_contents_information, 'undoLog')
	undoLog = PageTemplateFile('www/undoLog', globals())

	security.declareProtected(Permissions.access_contents_information, 'printView')
	printView = PageTemplateFile('www/printView', globals())

	security.declareProtected(Permissions.access_contents_information, 'advancedSearch')
	advancedSearch = PageTemplateFile('www/advancedSearch', globals())

	security.declareProtected(Permissions.access_contents_information, 'batchNavigation')
	batchNavigation = PageTemplateFile('www/batchNavigation', globals())

	security.declareProtected(Permissions.access_contents_information, 'details')
	details = PageTemplateFile('www/details', globals())

	security.declareProtected(Permissions.access_contents_information, 'editRow')
	editRow = PageTemplateFile('www/editRow', globals())

	# debug template
	security.declareProtected(Permissions.access_contents_information, 'debugTemplate')
	debugTemplate = PageTemplateFile('www/debugTemplate', globals())

## DTML methods
	frontend_css = DTMLFile('www/frontend.css', globals())
	print_css = DTMLFile('www/print.css', globals())
	javascripts_js = DTMLFile('www/javascripts.js', globals())
	datepicker_js = DTMLFile('www/datepicker.js', globals())
	table_reorder_js = DTMLFile('www/table_reorder.js', globals())
	nodomws_js = DTMLFile('www/nodomws.js', globals())

## ImageFile methods
	favicon_ico = ImageFile('www/favicon.ico', globals())
	details_gif = ImageFile('www/images/details.gif', globals())
	edit_gif = ImageFile('www/images/edit.gif', globals())
	duplicate_gif = ImageFile('www/images/duplicate.gif', globals())
	delete_gif = ImageFile('www/images/delete.gif', globals())
	undo_gif = ImageFile('www/images/undo.gif', globals())
	up_gif = ImageFile('www/images/up.gif', globals())
	down_gif = ImageFile('www/images/down.gif', globals())
	plus_gif = ImageFile('www/images/plus.gif', globals())
	minus_gif = ImageFile('www/images/minus.gif', globals())
	top_gif = ImageFile('www/images/top_c140.gif', globals())
	# calendar images
	cal_gif = ImageFile('www/images/cal.gif', globals())
	bg_header_jpg = ImageFile('www/images/bg_header.jpg', globals())
	gradient_gif = ImageFile('www/images/gradient-e5e5e5-ffffff.gif', globals())
	bullet1_gif = ImageFile('www/images/bullet1.gif', globals())
	bullet2_gif = ImageFile('www/images/bullet2.gif', globals())

## variable getter and setter methods
	# by default zope has a "deny unless allowed" policy, but Item and SimpleItem override this and make non protected methods available
	security.declareProtected(Permissions.view_management_screens,'getDbConnectionId')
	def getDbConnectionId(self):
		"""
		Returns the current dbConnectionId
		"""
		return self._dbConnectionId

	security.declareProtected(Permissions.view_management_screens,'setDbConnectionId')
	def setDbConnectionId(self, dbConnectionId):
		"""
		Sets the dbConnectionId and tries connecting to the db
		returns True if successfully connected, otherwise False
		"""
		self._dbConnectionId = dbConnectionId
		self._dbConnection = getattr(self, self._dbConnectionId, None)
		if self._dbConnection :
			return True
		return False

	security.declareProtected(Permissions.access_contents_information, 'getLastErrorMessage')
	def getLastErrorMessage(self) :
		"""
		Returns the error message from the last query or None if there was no error.
		"""
		return self._error

	security.declareProtected(Permissions.access_contents_information, 'getLogTable')
	def getLogTable(self) :
		"""
		Returns the table used for logging.
		"""
		return self._logTable

## helper methods
	security.declareProtected(Permissions.access_contents_information, 'databaseIsConnected')
	def databaseIsConnected(self) :
		"""
		Returns true if database is connected.
		"""
		# refresh connection in case it has changed
		self._dbConnection = getattr(self, self._dbConnectionId, None)
		if not self._dbConnection :
			return False
		return True

	security.declareProtected(Permissions.access_contents_information, 'testTableSelection')
	def testTableSelection(self) :
		"""
		Tests if a table is selected, and if not redirect to index page.
		This is used in all subpages to make redirection in case that no table is selected (e.g. if the session has expired).
		"""
		if not self.getSelectedTable() :
			self.REQUEST.RESPONSE.redirect(self.index_html.absolute_url())

	security.declareProtected(Permissions.view_management_screens,'getDbConnections')
	def getDbConnections(self) :
		"""
		Returns all available db connections
		"""
		#connections = self.get_container().objectValues('Z MySQL Database Connection')
		#connectionIds = [connection.getId() for connection in connections]
		#return connectionIds
		ids = SQLConnectionIDs(self)
		return [x[1] for x in ids]

	security.declareProtected(Permissions.view_management_screens, 'getRoles')
	def getRoles(self) :
		"""
		return roles dictionary
		"""
		return self._roles

	security.declareProtected(Permissions.access_contents_information, 'getUsername')
	def getUsername(self) :
		"""
		return the name of the logged in user
		"""
		return getSecurityManager().getUser().getUserName()

	security.declareProtected(Permissions.access_contents_information, 'userIsAdmin')
	def userIsAdmin(self) :
		"""
		Returns the admin privileges of the current user.
		possible roles:
			0 -> only view rights (use silva access settings to control view rights)
			1 -> reserved for view rights (now implemented via silva access settings)
			2 -> view secrete columns
			3 -> edit data
			4 -> add/delete data
		TODO:
		- add role settings to manage_edit_view
		"""
		possibleRoles = ['Viewer','Viewer +','Viewer ++','Author','Editor','ChiefEditor','Manager']
		allRoles = []

		user = getSecurityManager().getUser().getId()

                #roles = self.sec_get_all_roles_for_userid(user)
		#roles = self.sec_get_all_roles(user)
                roles = self.sec_get_all_roles()

		if 'Manager' in roles :
			allRoles = possibleRoles
		elif 'ChiefEditor' in roles :
			allRoles = possibleRoles[:6]
		elif 'Editor' in roles :
			allRoles = possibleRoles[:5]
		elif 'Author' in roles :
			allRoles = possibleRoles[:4]
		elif 'Viewer ++' in roles :
			allRoles = possibleRoles[:3]
		elif 'Viewer +' in roles :
			allRoles = possibleRoles[:2]
		elif 'Viewer' in roles :
			allRoles = possibleRoles[:1]

		if self._roles :
			if self._roles.get('addRemoveData') in allRoles or not self._roles.has_key('addRemoveData') :
				return 4
			elif self._roles.get('editData') in allRoles or not self._roles.has_key('editData') :
				return 3
			elif self._roles.get('viewSecretColumns') in allRoles or not self._roles.has_key('viewSecretColumns') :
				return 2
			else :
				return 0
		else :
		# no roles defined -> all users have all rights
			return 4

	security.declareProtected(Permissions.access_contents_information, 'userIsManager')
	def userIsManager(self) :
		"""
		Returns True if user has Manager rights, otherwise False
		"""
		user = getSecurityManager().getUser().getId()
                #roles = self.sec_get_all_roles_for_userid(user)
		#roles = self.sec_get_all_roles(user)
                roles = self.sec_get_all_roles()
		if 'Manager' in roles :
			return True
		return False

	security.declareProtected(Permissions.access_contents_information, 'userIsAnonymous')
	def userIsAnonymous(self) :
		"""
		Returns True if user is Anonymous
		"""
		if getSecurityManager().getUser().getUserName() == 'Anonymous User' :
			return True
		return False

	security.declareProtected(Permissions.access_contents_information, 'processScript')
	def processScript(self) :
		"""
		Currently does only deleting of rows (is called from index page).
		"""
		delete = getattr(self.REQUEST,'delete',None)

		if delete == 'yes' :
		# delete row
			if self.selectedTableIsWritable() and self.userIsAdmin() >= 4 :
				self.deleteRow()
				return (0,'Record deleted!')
			else :
				return (1,'You do not have the privileges to delete!')

		return None

	security.declareProtected(Permissions.access_contents_information, 'processUndo')
	def processUndo(self) :
		"""
		Performs the undo operation (called from undoLog).
		"""
		undoID = getattr(self.REQUEST, 'undo', None)
		emptyLog = getattr(self.REQUEST, 'emptyLog', None)

		if undoID :
			if self.userIsAdmin() < 4 :
				return (1,'You do not have the privileges to undo actions!')
			# get undo query
			query = "SELECT `sql_query`, `undo_sql_query`, `timestamp`, `database`, `table`, `record_key` FROM %s WHERE `id`=%s" % (self._logTable, undoID)
			result = self.queryString(query)[0]
			query = result['sql_query']
			undoQuery = result['undo_sql_query']
			timestamp = result['timestamp']
			database = result['database']
			table = result['table']
			record_key = result['record_key']
			values = {'database':database, 'table':table, 'record_key':record_key}
			self.queryString(undoQuery)
			if self.getLastErrorMessage() :
				warning = 1
				message = "Could not undo action. MySQL said: %s" % self.getLastErrorMessage()
			else :
				self.logAction(action='undo', query=undoQuery, values=values, timestamp=timestamp, undoQuery=query)
				warning = 0
				message = "Successfully undone action."
			return (warning, message)
		elif emptyLog == 'yes' :
			if self.userIsManager() :
				query = "DELETE FROM %s" % self._logTable
				self.queryString(query)
				return (0, "Log successfully emptied!")
			else :
				return (1, "You do not have the privileges to empty the log!")
		else :
			return None

	security.declareProtected(Permissions.access_contents_information, 'getColumns')
	def getColumns(self) :
		"""
		Returns columns for use in SELECT statements.
		"""
		columns = self.getColumnsAsArray()
		keys = []

		for key in self.getPrimaryKeys() :
			if key not in columns :
				keys.append(key)

		return '`%s`' % '`,`'.join(columns+keys)

	security.declareProtected(Permissions.access_contents_information, 'getDetailsColumns')
	def getDetailsColumns(self) :
		"""
		Returns columns for detail view for use in SELECT statements.
		"""
		columns = [x[0] for x in self.getFields()]
		return '`%s`' % '`,`'.join(columns)

	security.declareProtected(Permissions.access_contents_information, 'hiddenColumns')
	def hiddenColumns(self, view='list') :
		"""
		Returns all hidden columns for the selected table and view (currently hidden columns exist only for list view).
		"""
		tables = self._columnSettings.get(self.getSelectedDatabase(),{})
		views = tables.get(self.getSelectedTable(),{})
		if views.has_key('secretColumns') :
		# _secretColumns is in new format
			fields = views.get(view,{})
		else :
		# _secretColumns is in old format (<= v0.6(beta3))
			fields = views
		hiddenColumns = []
		for field in fields :
			if fields[field].get('hidden') :
				hiddenColumns.append(field)
		return hiddenColumns

	security.declareProtected(Permissions.access_contents_information, 'secretColumns')
	def secretColumns(self) :
		"""
		Returns all secret columns for the selected table.
		"""
		tables = self._columnSettings.get(self.getSelectedDatabase(),{})
		views = tables.get(self.getSelectedTable(),{})
		if not views.has_key('secretColumns') :
		# _secretColumns is in old format (<= v0.6(beta3))
			secretColumns = []
			for field in views :
				if views[field].get('secret') :
					secretColumns.append(field)
		else :
		# _secretColumns is in new format
			secretColumns = views['secretColumns']
		return secretColumns

	security.declareProtected(Permissions.access_contents_information, 'mapping')
	def mapping(self, column, view='list') :
		"""
		Returns the mapped name for a given column and view.
		"""
		tables = self._columnSettings.get(self.getSelectedDatabase(),{})
		views = tables.get(self.getSelectedTable(),{})
		if views.has_key('secretColumns') :
		# _secretColumns is in new format
			fields = views.get(view,{})
			if view == 'details' and not fields :
			# fall back to list settings
				fields = views.get('list',{})
		else :
		# _secretColumns is in old format (<= v0.6(beta3))
			fields = views
		mapping = None
		if fields.has_key(column) :
			mapping = fields[column].get('mapping')
		if not mapping :
			# set mapping to column name if empty string or not set
			mapping = column
		return mapping

	security.declareProtected(Permissions.access_contents_information, 'columnHelp')
	def columnHelp(self, column, view='details') :
		"""
		Returns the help for a given column and view (currently column help exists only for details view).
		"""
		tables = self._columnSettings.get(self.getSelectedDatabase(),{})
		views = tables.get(self.getSelectedTable(),{})
		if views.has_key('secretColumns') :
		# _secretColumns is in new format
			fields = views.get(view,{})
		else :
		# _secretColumns is in old format (<= v0.6(beta3))
			fields = views
		if fields.has_key(column) :
			return fields[column].get('help')
		else :
			return None

	security.declareProtected(Permissions.access_contents_information, 'columnTitle')
	def columnTitle(self, column, view='details') :
		"""
		Returns the help for a given column and view (currently column titles exist only for details view).
		"""
		tables = self._columnSettings.get(self.getSelectedDatabase(),{})
		views = tables.get(self.getSelectedTable(),{})
		if views.has_key('secretColumns') :
		# _secretColumns is in new format
			fields = views.get(view,{})
		else :
		# _secretColumns is in old format (<= v0.6(beta3))
			fields = views
		if fields.has_key(column) :
			return fields[column].get('title')
		else :
			return None

	security.declareProtected(Permissions.access_contents_information, 'groupFieldsByTitle')
	def groupFieldsByTitle(self, fields) :
		"""
		Groups the given fields into an array by the saved column titles.
		E.g.: array( {title:'', fields:[a,b,c]}, {title:'Options', fields:[c,d,e,f,g]}, {title:'Extras', fields:[x,y,z]} )
		"""
		grouped = []
		ff = []
		title = ''
		for field in fields :
			if self.columnTitle(field) :
			# new column title -> add group and empty array
				if ff :
				# if first item has already a title, ff is empty
					grouped.append({'title':title, 'fields':ff})
					ff = []
				title = self.columnTitle(field) or ''
			ff.append(field)
		if ff :
		# append last group
			grouped.append({'title':title, 'fields':ff})
		return grouped

	security.declareProtected(Permissions.access_contents_information, 'showTableSelector')
	def showTableSelector(self) :
		"""
		Returns True if table selector should be shown.
		"""
		return self._showTableSelector

	security.declareProtected(Permissions.access_contents_information, 'nullString')
	def nullString(self) :
		"""
		Returns the string to be used for NULL values.
		"""
		return self._nullString

	security.declareProtected(Permissions.access_contents_information, 'footer')
	def footer(self) :
		"""
		Returns the footer to be displayed at the bottom of every page.
		"""
		return self._footer

	security.declareProtected(Permissions.access_contents_information, 'getDetailsWhere')
	def getDetailsWhere(self) :
		"""
		Returns the WHERE statement for the detail view.
		"""
		req = self.REQUEST.form
		keys = self.getPrimaryKeys()
		args = []

		for key in keys :
		  if req.has_key(key) :
			#args.append('`'+key+'`="'+req[key]+'"')
			args.append('`%s`=%s' % (key, self.escape(req[key])))
		  else :
			args.append('`'+key+'` IS NULL')

		return ' AND '.join(args)

	security.declareProtected(Permissions.access_contents_information, 'getUrlForRow')
	def getUrlForRow(self, row) :
		"""
		Returns the url for a given row.
		"""
		keys = self.getPrimaryKeys()

		url = []
		for key in keys :
		  if row[key]!=None :
		  	if isinstance(row[key],array) :
		  		url.append(key+'='+self.stringFromBlob(row[key]))
		  	else :
				url.append(key+'='+str(row[key]))
		return '?'+'&'.join(url)

	security.declareProtected(Permissions.access_contents_information, 'getValues')
	def getValues(self) :
		"""
		Returns the values taken from the REQUEST (used when editing rows).
		"""
		if self.REQUEST.get('submitNow', None):
			fields = self.REQUEST.form
			values = {}
			for field in fields :
				if field[:7] == 'fields_' :
					values[field[7:]] = fields[field]
		else :
                        action = self.REQUEST.get('action', None)
			if action in ['edit',] :
			# save where argument in session to identify row later
				self.REQUEST.SESSION['editWhere'] = self.getDetailsWhere()
			# get values
			if action in ['edit', 'duplicate',] :
				values = self.getRow()
				if not values :
					return {}
				else :
					values = values.dictionaries()
			else :
				values = {}
			if values :
				values = values[0]
			else :
				values = {}
				for field in self.getFields() :
					if field[2] == 'CURRENT_TIMESTAMP' :
						values[field[0]] = None
					else :
						values[field[0]] = field[2]
			if action in ['add', 'duplicate',] :
			# set unique keys to NULL and auto_increment fields to next auto_increment value
				uniqueKeys = self.getUniqueKeys()
				nextID = self.getTableStatus()[0]['Auto_increment']
				for value in values :
					if value in uniqueKeys :
						if uniqueKeys[value] == 'auto_increment' :
							values[value] = nextID
						else :
							values[value] = None
		return values

	security.declareProtected(Permissions.access_contents_information, 'htmlEncode')
	def htmlEncode(self, string) :
		"""
		Returns the given string html encoded.
		"""
		return newline_to_br(html_quote(string))

	security.declarePrivate('columnSort')
	def columnSort(self,x,y,view) :
		"""
		Sort function for sorting by order saved in columnSettings
		"""
		# sort fields
		tables = self._columnSettings.get(self.getSelectedDatabase(),{})
		views = tables.get(self.getSelectedTable(),{})
		if views.has_key('secretColumns') :
		# _secretColumns is in new format
			columns = views.get(view,{})
			if view == 'details' and not columns :
			# fall back to list settings
				columns = views.get('list',{})
		else :
		# _secretColumns is in old format (<= v0.6(beta3))
			columns = views
		if not columns :
		# no column settings saved
			return 0
		if not columns.get(x) or columns[x].get('order') == None :
		# no order saved for x
			return 1
		if not columns.get(y) or columns[y].get('order') == None :
		# no order saved for y
			return -1
		return cmp(columns[x]['order'],columns[y]['order'])

	security.declareProtected(Permissions.access_contents_information, 'writeRowToDB')
	def writeRowToDB(self) :
		"""
		Writes the passed row to the db. Checks if user has enough rights to do this.
		"""
		action = getattr(self.REQUEST,'action',None)
		session = self.REQUEST.SESSION
		editAgain = getattr(self.REQUEST,'editAgain',None)

		if self.selectedTableIsWritable() and ((self.userIsAdmin() >= 4 and action == 'add') or (self.userIsAdmin() >= 3 and action == 'edit')) :
		# get fields and values
			#fields = [x[0] for x in self.getFields() if x[1]!='blob']
			fields = [x[0] for x in self.getFields() if not self.fieldIsBlob(x[0])]
			values = [getattr(self.REQUEST,'fields_'+x,'NULL') for x in fields]
			for i in range(len(values)) :
				if values[i] == 'NULL' :
				# if user wrote NULL
					values[i] = None
#				if same_type(values[i],[]) :
				if isinstance(values[i],list) :
					values[i] = ','.join(values[i])
			if action == 'add' :
			# add row
				# values[:] to use a copy, otherwise the original will be escaped
				self.addRow(fields=fields,values=values[:])
				if self.getLastErrorMessage() :
					message = 'There was an error, the record has not been added!\nMySQL said: '+self.getLastErrorMessage()
					warning = 1
				else :
					message = 'Record added!'
					warning = 0
			if action == 'edit' :
			# edit row
				# values[:] to use a copy, otherwise the original will be escaped
				self.updateRow(fields=fields,values=values[:],where=session['editWhere'])
				if self.getLastErrorMessage() :
					message = 'There was an error, the record has not been updated!\nMySQL said: '+self.getLastErrorMessage()
					warning = 1
				else :
					message = 'Record updated!'
					warning = 0
					if not editAgain :
						del session['editWhere']
			if warning :
				return self.editRow(warning=(warning,message))
			else :
				if editAgain :
					if action == 'add' :
						# set action to edit
						self.REQUEST.set('action','edit')
					# set editwhere (in case that it is a new row or that the primary key has changed)
					keys = self.getPrimaryKeys()
					args = []
					for key in keys :
						#value = values[fields.index(key)]
						value = getattr(self.REQUEST,'fields_'+key,None)
						if value != None :
							#args.append('`'+key+'`="'+value+'"')
							args.append('`%s`=%s' % (key, self.escape(value)))
						else :
							args.append('`'+key+'` IS NULL')
					self.REQUEST.SESSION['editWhere'] = ' AND '.join(args)
					return self.editRow(warning=(warning,message))
				else :
					return self.index_html(warning=(warning,message))
		elif action :
			return self.index_html(warning=(1,'You do not have the privileges to add or update!'))
		# nothing to write to db...
		return self.index_html()

	security.declareProtected(Permissions.access_contents_information, 'exportAsCSV')
	def exportAsCSV(self) :
		"""
		Exports the current result as CSV.
		"""
		request  = self.REQUEST
		response = request.RESPONSE
		session = request.SESSION
		type = getattr(request,'type','')
		encoding = getattr(request,'encoding',None)
		csv = ''

		# save type and encoding in session
		session['type'] = type
		session['encoding'] = encoding

		def makeunicode(text, encoding ='utf8'):
        # make unicode
			try:
				return unicode(text, encoding, 'replace')
			except:
				return text

		def escapeValue(value,type) :
		# escapes strings according to passed type
			if value :
				value = str(value)
				value = value.replace('\r\n',' ')
				value = value.replace('\r',' ')
				value = value.replace('\n',' ')
				if type == 'excel' :
					if value.count('"') or value.count(',') :
						value = value.replace('"','""')
						value = '"'+value+'"'
				else :
					value = value.replace('\t',' ')
			else :
				value = ''
			return value

		def generateLine(values,type) :
		# generate a line from array values according to passed type
			if type == 'excel' :
				line = ','.join(values)+'\n'
			else :
				line = '\t'.join(values)+'\n'
			return line

		# generate csv file
		## generate headers
		fields = self.getColumnsAsArray()
		ff = []
		for field in fields :
			field = self.mapping(field)
			ff.append(escapeValue(field,type))
		csv += generateLine(ff,type)
		## generate rows
		rows = self.getResult().dictionaries()
		for row in rows :
			values = []
			for field in fields :
				values.append(escapeValue(row[field],type))
			csv += generateLine(values,type)

		## encode
		if encoding :
			csv = makeunicode(csv).encode(encoding,'replace')


		# send csv file
		## get filename
		fileName = self.getSelectedDatabase()+'_'+self.getSelectedTable()
		if type == 'excel':
			fileName += '.csv'
		else :
			fileName += '.txt'
		## send file
		response.setHeader('Content-type', 'octet-stream/octet-stream')
		response.setHeader('Content-disposition', 'inline; filename="'+fileName+'"')
		return csv

	security.declareProtected(Permissions.access_contents_information, 'getPermalink')
	def getPermalink(self, page):
		"""
		Returns a link which has all session variables as GET string and links to page
		"""
		req = self.REQUEST
		session = req.SESSION
		vars = {}

		# get session values
		sessionVariables = {}
		for var in session.keys() :
			if var in ['selectedDatabase','selectedTable','selectedColumns','b_start','where','sort', 'batchSize','encoding','type'] :
				sessionVariables[var] = session[var]
		vars['sessionVariables'] = marshal.dumps(sessionVariables)

		# get get/post values
		for var in req.form.keys() :
			vars[var] = req.form[var]

		return '%s?%s' % (page.absolute_url(), urlencode(vars) )

	security.declareProtected(Permissions.access_contents_information, 'getSessionTimeout')
	def getSessionTimeout(self) :
		"""
		returns the zope session timeout in minutes
		"""
		sess = self.temp_folder.session_data
		return sess.getTimeoutMinutes()

	security.declarePrivate('setSessionTimeout')
	def setSessionTimeout(self,minutes) :
		"""
		sets the zope session timeout to minutes
		returns the zope session timeout in minutes
		"""
		sess = self.temp_folder.session_data
		timeout=int(minutes)
		if self.getSessionTimeout() != timeout:
			sess.setTimeoutMinutes(timeout)
		return self.getSessionTimeout()

	security.declareProtected(Permissions.access_contents_information, 'getCookies')
	def getCookies(self) :
		"""
		returns all cookies
		"""
		cookies = self.REQUEST.cookies

		cc = [x[12:] for x in cookies if x.startswith('sessionSave_')]

		return cc

	security.declareProtected(Permissions.access_contents_information, 'saveSessionToCookie')
	def saveSessionToCookie(self,name):
		"""
		saves the current session in a cookie with the given name.
		"""
		request = self.REQUEST
		session = request.SESSION
		response = request.RESPONSE
		name = 'sessionSave_%s' % name

		# get session values
		sessionVariables = {}
		for var in session.keys() :
			if var in ['selectedDatabase','selectedTable','selectedColumns','b_start','where','sort', 'batchSize','encoding','type'] :
				sessionVariables[var] = session[var]

		# remove whitespaces in name (would break cookies/request)
		name = ''.join(name.split())

		vars = marshal.dumps(sessionVariables)
		# encode string to avoid breaking the cookies (last char would be a \n so we delete it)
		vars = base64.encodestring(vars)[:-1]
		# on long strings, the base64 encoding can contain multiple lines separated by \n or \r\n which should be deleted
		vars = ''.join(vars.split())

		# TODO: use right date
		response.setCookie(name, vars, expires='Wed, 19 Feb 2020 14:28:00 GMT')

		# add it to cookies array (otherwise it won't be in the cookies array before the next page load)
		request.cookies[name] = vars

	security.declareProtected(Permissions.access_contents_information, 'loadCookie')
	def loadCookie(self,name) :
		"""
		loads the cookie with the given name
		"""
		request = self.REQUEST
		cookies = request.cookies
		session = request.SESSION
		name = 'sessionSave_%s' % name

		if not cookies.has_key(name) :
			return

		# delete old keys
		keys = ['selectedDatabase','selectedTable','primaryKeys','uniqueKeys','selectedColumns','b_start','where','sort','fields','blobs']
		for key in keys :
			if session.has_key(key) :
				del session[key]
		# load new keys from cookie
		sessionVariables = marshal.loads(base64.decodestring(cookies[name]))
		for key in sessionVariables :
			session[key] = sessionVariables[key]

	security.declareProtected(Permissions.access_contents_information, 'deleteCookie')
	def deleteCookie(self,name) :
		"""
		deletes the specified cookie
		"""
		request = self.REQUEST
		cookies = request.cookies

		name = 'sessionSave_%s' % name
		request.RESPONSE.setCookie(name,'',expires="Thu, 01-Jan-70 00:00:01 GMT")
		if cookies.has_key(name) :
			del cookies[name]


## MySQL methods
	security.declarePrivate('queryString')
	def queryString(self, query, error=True) :
		"""
		Performs the given string as MySQL query.
		If error is False, MySQL errors are not save to _error (i.e. getLastErrorMessage() will still return the error from the last MySQL query).
		"""
		# refresh connection in case it has changed
		self._dbConnection = getattr(self, self._dbConnectionId, None)
		# set encoding to utf8
		proc_id = "utf8_proc"
		sql_method = SQL(proc_id, '', self._dbConnectionId, {}, 'SET NAMES "utf8"')
		setattr(sql_method, self._dbConnectionId, self._dbConnection)
		sql_method(REQUEST={})
		# perform query
		proc_id = "query_proc"
		sql_method = SQL(proc_id, '', self._dbConnectionId, {}, query)
		# disable result limit
		sql_method.max_rows = sql_method.max_rows_ = 0
		setattr(sql_method, self._dbConnectionId, self._dbConnection)
		try :
			result = sql_method(REQUEST={})
			if error :
				self._error = None
		except Exception,e :
			result = []
			if error :
				self._error = str(e)+' query was: '+query
		return result

	security.declarePrivate('directQueryString')
	def directQueryString(self, query) :
		"""
		Performs a query directly using the connection object, not using an SQL Method object.
		Currently wrong result format and no error support.
		"""
		self._dbConnection = getattr(self, self._dbConnectionId, None)
		conn = self._dbConnection()
		return conn.query(query)

	security.declarePrivate('escape')
	def escape(self, object) :
		"""
		Escapes a string/int or array/dictionary for use in MySQL.
		object can be a string/int or an array/dictionary. If it is an array/dictionary,
		all elements in it are escaped. If a value is None, it is replaced by NULL.
		Elements are also quoted (except NULL).
		"""
		if object == None :
			object = 'NULL'
		elif isinstance(object,str) or isinstance(object,int) :
			object = escape_string(str(object))
			object = '"%s"' % object
		else :
			for i in xrange(len(object)) :
				if object[i] == None :
					object[i] = 'NULL'
				else :
					object[i] = '"%s"' % escape_string(str(object[i]))
#			object = [x and '"%s"' % str(x) or 'NULL' for x in object]
		return object

	security.declareProtected(Permissions.access_contents_information, 'stringFromBlob')
	def stringFromBlob(self, blob) :
		"""
		returns a string from the given blob (works also if a string or int is passed)
		"""
		if isinstance(blob,array) :
			return blob.tostring()
		else :
			return str(blob)

	security.declareProtected(Permissions.access_contents_information, 'checkLogTable')
	def checkLogTable(self, table) :
		"""
		Returns true if table has correct structure for use with logging.
		"""
		result = self.queryString("DESCRIBE %s" % table)
		if len(result) < 6 :
			return False
		x = 0
		for row in result :
			if x == 0 and (not row['Type'].startswith('int') or row['Key']!='PRI' or row['Extra'] != 'auto_increment') :
			# id
				return False
			if x == 1 and row['Type'] != 'timestamp' :
			# timestamp
				return False
			if x == 2 and (not row['Type'].startswith('varchar') and not row['Type'] == 'text') :
			# user
				return False
			if x == 3 and row['Type'] != "enum('insert','update','delete','undo')" :
			# action_type
				return False
			if x == 4 and (not row['Type'].startswith('varchar') and not row['Type'] == 'text') :
			# database
				return False
			if x == 5 and (not row['Type'].startswith('varchar') and not row['Type'] == 'text') :
			# table
				return False
			if x == 6 and (not row['Type'].startswith('text') and not row['Type'].startswith('varchar') and not row['Type'] == 'int') :
			# record_key
				return False
			if x == 7 and row['Type'] != 'timestamp' :
			# undo_from
				return False
			if x == 8 and (not row['Type'].startswith('varchar') and not row['Type'] == 'text') :
			# sql_query
				return False
			if x == 9 and (not row['Type'].startswith('varchar') and not row['Type'] == 'text') :
			# undo_sql_query
				return False
			x += 1
		return True

	security.declarePrivate('logAction')
	def logAction(self, action, query, values, timestamp=None, undoQuery=None) :
		"""
		Logs the passed action/query/values to the logging table (if defined).
		action: the action (insert/update/delete/undo)
		query: the performed query
		values: the old values in case of a delete/update, the new values in case of an insert, database/table/key in case of an undo
		timestamp: the timestamp of the undone action in case of an undo
		undoQuery: the query necessary to undo the action (only passed in case of an undo)
		"""
		if not self._logTable :
		# no log table defined
			return
		if action == 'undo' :
			selectedDatabase = values['database']
			selectedTable = values['table']
			recordKey = values['record_key']
			undoFrom = str(timestamp)
		else :
			selectedDatabase = self.getSelectedDatabase()
			selectedTable = self.getSelectedTable()
			keys = self.getPrimaryKeys()
			recordKey = ''
			for key in keys :
				if recordKey :
					recordKey += '|'
				if values.has_key(key) and values[key]!=None :
					if isinstance(values[key],array) :
						recordKey += self.stringFromBlob(values[key])
					else :
						recordKey += str(values[key])
				else :
					recordKey += 'NULL'
			undoFrom = ''
		if action == 'insert' :
		# get delete query
			args = []
			keys = self.getPrimaryKeys()
			for key in keys :
			  if values.has_key(key) :
				args.append('`%s`=%s' % (key, self.escape(values[key])))
			  else :
				args.append('`'+key+'` IS NULL')
			undoQuery = 'DELETE FROM `%s`.`%s` WHERE %s LIMIT 1' % (self.getSelectedDatabase(), self.getSelectedTable(), ' AND '.join(args))
		elif action == 'update' or action == 'delete' :
		# get replace into query
			for key in values :
			# fix blob values
				if isinstance(values[key],array) :
					values[key] = self.stringFromBlob(values[key])
			fields = '`%s`' % '`, `'.join(values.keys())
			values = ', '.join(self.escape(values.values()))
			undoQuery = 'REPLACE INTO `%s`.`%s` (%s) VALUES (%s)' % (self.getSelectedDatabase(), self.getSelectedTable(), fields, values)
		query = 'INSERT INTO %s VALUES (NULL, NULL, %s, %s, %s, %s, %s, %s, %s, %s)' % (self._logTable, self.escape(self.getUsername()), self.escape(action), self.escape(selectedDatabase), self.escape(selectedTable), self.escape(recordKey), self.escape(undoFrom), self.escape(query), self.escape(undoQuery))
		self.queryString(query, False)

	security.declareProtected(Permissions.access_contents_information, 'getLog')
	def getLog(self) :
		"""
		Returns the log.
		"""
		if self.userIsAdmin() < 4 :
			return []
		query = "SELECT * FROM %s ORDER BY `timestamp` DESC" % self._logTable
		return self.queryString(query)

	security.declareProtected(Permissions.access_contents_information, 'getDatabases')
	def getDatabases(self) :
		"""
		Returns all databases.
		"""
		if not self.databaseIsConnected() :
			return []
		result = self.queryString("SHOW DATABASES")
		databases = []
		for entry in result :
			databases.append(entry[0])
		return databases

	security.declareProtected(Permissions.access_contents_information, 'getTables')
	def getTables(self, db=None) :
		"""
		Returns all tables in the given database.
		If no database is specified, the selected database (from the session) is used.
		"""
		if not db :
			db = self.getSelectedDatabase()
		tables = []
		if db :
			result = self.queryString("SHOW TABLES FROM `%s`" % db)
			#result = self.queryString('SHOW TABLES FROM <dtml-sqlvar "\'zope\'" type=int>')
			for entry in result :
				tables.append(entry[0])
		return tables

	security.declarePrivate('addRow')
	def addRow(self, fields, values) :
		"""
		Adds a row to the selected database and table.
		"""
		# add row
		ff = '`%s`' % '`, `'.join(fields)
		vv = ', '.join(self.escape(values[:]))
		query = "INSERT INTO `%s`.`%s` (%s) VALUES (%s)" % (self.getSelectedDatabase(), self.getSelectedTable(), ff, vv)
		self.queryString(query)
		# log action
		if not self.getLastErrorMessage() :
			values = dict(zip(fields,values))
			self.logAction('insert',query,values)

	security.declarePrivate('updateRow')
	def updateRow(self, fields, values, where) :
		"""
		Updates the selected row.
		"""
		# get old values
		query = "SELECT * FROM `%s`.`%s` WHERE %s" % (self.getSelectedDatabase(), self.getSelectedTable(), where)
		result = self.queryString(query)
		oldValues = result.dictionaries()[0]
		# update row
		vv = self.escape(values[:])
		sets = []
		for i in xrange(len(fields)) :
			sets.append("`%s`=%s" % (fields[i],vv[i]))
		set = ', '.join(sets)
		query = "UPDATE `%s`.`%s` SET %s WHERE %s LIMIT 1" % (self.getSelectedDatabase(), self.getSelectedTable(), set, where)
		self.queryString(query)
		# log action
		if not self.getLastErrorMessage() :
			self.logAction('update',query,oldValues)

	security.declarePrivate('deleteRow')
	def deleteRow(self) :
		"""
		Deletes a row.
		"""
		where = self.getDetailsWhere()
		# get old values
		query = "SELECT * FROM `%s`.`%s` WHERE %s" % (self.getSelectedDatabase(), self.getSelectedTable(), where)
		result = self.queryString(query)
		oldValues = result.dictionaries()[0]
		# delete row
		query = "DELETE FROM `%s`.`%s` WHERE %s LIMIT 1" % (self.getSelectedDatabase(), self.getSelectedTable(), where)
		self.queryString(query)
		# log action
		if not self.getLastErrorMessage() :
			self.logAction('delete',query, oldValues)

	security.declareProtected(Permissions.access_contents_information, 'describeTable')
	def describeTable(self, db=None, table=None) :
		"""
		Get information about a table (DESCRIBE).
		Uses the given db and table or if not specified the db and table from the session.
		"""
		if not db :
			db = self.getSelectedDatabase()
		if not table :
			table = self.getSelectedTable()
		query = "DESCRIBE `%s`.`%s`" % (db, table)
		return self.queryString(query)

	security.declareProtected(Permissions.access_contents_information, 'getTableStatus')
	def getTableStatus(self) :
		"""
		Get the status of the selected table.
		"""
		query = 'SHOW TABLE STATUS FROM `%s` LIKE "%s"' % (self.getSelectedDatabase(), self.getSelectedTable())
		return self.queryString(query)

	security.declareProtected(Permissions.access_contents_information, 'getResult')
	def getResult(self) :
		"""
		Get the rows for the selected database and table (applying search and sort arguments if given).
		"""
		if self.getSelectedTable() :
			query = "SELECT %s FROM `%s`.`%s`" % (self.getColumns(), self.getSelectedDatabase(), self.getSelectedTable())
			where = self.getWhere()
			if where :
				args = []
				for w in where :
					if isinstance(w,str) :
						args.append(w)
					else :
						if w['value'] != 'NULL' :
							arg = "`%s` %s %s" % (w['column'], w['operation'], self.escape(w['value']))
						elif w['operation'] == '!=' :
							arg = "`%s` IS NOT NULL" % w['column']
						else :
							arg = "`%s` IS NULL" % w['column']
						args.append(arg)
				query += " WHERE %s" % ' AND '.join(args)
			sort = self.getSort()
			if sort :
				query += " ORDER BY `%s` %s" % (sort[0], sort[1])
			return self.queryString(query)
		else :
			return []

	security.declareProtected(Permissions.access_contents_information, 'getRow')
	def getRow(self) :
		"""
		Get the selected row.
		"""
		query = "SELECT %s FROM `%s`.`%s` WHERE %s" % (self.getDetailsColumns(), self.getSelectedDatabase(), self.getSelectedTable(), self.getDetailsWhere())
		return self.queryString(query)

	security.declareProtected(Permissions.access_contents_information, 'numberOfRows')
	def numberOfRows(self) :
		"""
		Get the number of rows in the selected table.
		"""
		if self.getSelectedTable() :
			query = "SELECT COUNT(*) AS numRows FROM `%s`.`%s`" % (self.getSelectedDatabase(), self.getSelectedTable())
			result = self.queryString(query)
			return result[0]['numRows']
		return None

## session methods
	security.declareProtected(Permissions.access_contents_information, 'getSelectedDatabase')
	def getSelectedDatabase(self):
		"""
		Returns selected database.
		"""
		session = self.REQUEST.SESSION

		if session.has_key('selectedDatabase') :
		# return db from session
			return session['selectedDatabase']
		else :
		# no db selected
			if self.getPreselectTable() :
				return self.getPreselectTable()[0]
			return None

	security.declareProtected(Permissions.access_contents_information, 'getSelectedTable')
	def getSelectedTable(self):
		"""
		Returns selected table.
		"""
		session = self.REQUEST.SESSION

		if session.has_key('selectedTable') :
		# return table from session
			return session['selectedTable']
		else :
		# no table selected
			if self.getPreselectTable() and not session.has_key('selectedDatabase') :
				# only return preselect table if no db selected
				return self.getPreselectTable()[1]
			return None

	security.declareProtected(Permissions.access_contents_information, 'getPreselectTable')
	def getPreselectTable(self):
		"""
		Returns the database and table to be preselected (as a tuple "(database,table)").
		"""
		return self._preselectTable

	security.declareProtected(Permissions.access_contents_information, 'selectedTableIsWritable')
	def selectedTableIsWritable(self):
		"""
		Returns True if current table is writable and False if it is read-only.
		Currently only the logging table is read-only.
		"""
		table = "`%s`.`%s`" % (self.getSelectedDatabase(), self.getSelectedTable())
		if table == self.getLogTable() :
			return False
		return True

	security.declareProtected(Permissions.access_contents_information, 'getWhere')
	def getWhere(self) :
		"""
		Returns the WHERE statement for the current result.
		"""
		request = self.REQUEST
		session = request.SESSION

		if session.has_key('where') :
		# return where from session
		  return session['where']
		else :
		# no search selected
		  return None

	security.declareProtected(Permissions.access_contents_information, 'getSort')
	def getSort(self):
		"""
		Returns the sort criteria.
		"""
		request = self.REQUEST
		session = request.SESSION

		if session.has_key('sort') :
		# return sort from session
		  return session['sort']
		else :
		# no sort selected
		  return None

	security.declareProtected(Permissions.access_contents_information, 'fieldIsBlob')
	def fieldIsBlob(self,field):
		"""
		Returns True if field is blob, otherwise False.
		"""
		session = self.REQUEST.SESSION

		if not session.has_key('blobs') :
			self.getFields()

		return field in session['blobs']

	security.declareProtected(Permissions.access_contents_information, 'getBatchSize')
	def getBatchSize(self):
		"""
		Returns the current size of the batch.
		"""
		request = self.REQUEST
		session = request.SESSION

		if session.has_key('batchSize') :
		# return batch size from session
			return session['batchSize']
		else :
		# no batch size specified -> return default
			return 20

	security.declareProtected(Permissions.access_contents_information, 'getBlobs')
	def getBlobs(self):
		"""
		Returns all blob fields.
		"""
		session = self.REQUEST.SESSION

		if not session.has_key('blobs') :
			self.getFields()
		return session['blobs']

	security.declareProtected(Permissions.access_contents_information, 'getColumnsAsArray')
	def getColumnsAsArray(self, view='list'):
		"""
		Returns all fields (except hidden columns, secret columns only if enough privs) as array
		for the given view (database and table taken from session).
		Caches selected columns in session.
		"""
		request = self.REQUEST
		session = request.SESSION

		if session.has_key('selectedColumns') and session['selectedColumns'].has_key(view) :
			cols = session['selectedColumns'][view]
			if self.userIsAdmin() < 2 :
				for col in self.secretColumns() :
					if col in cols :
						cols.remove(col)
			return cols
		else :
			cols = [row['Field'] for row in self.describeTable()]
			for col in self.hiddenColumns() :
				if col in cols :
					cols.remove(col)
			if self.userIsAdmin() < 2 :
				for col in self.secretColumns() :
					if col in cols :
						cols.remove(col)
			cols.sort(lambda x,y : self.columnSort(x,y,view))
			if not session.has_key('selectedColumns') :
				session['selectedColumns'] = {}
			session['selectedColumns'][view] = cols
			return cols

	security.declareProtected(Permissions.view_management_screens, 'getFieldsForTable')
	def getFieldsForTable(self, db, table, view='list'):
		"""
		Returns all fields as array for the given db, table and view.
		"""
		fields = [row['Field'] for row in self.describeTable(db, table)]
		fields.sort(lambda x,y : self.columnSort(x,y,view))
		return fields

	security.declareProtected(Permissions.access_contents_information, 'getFields')
	def getFields(self, view='list'):
		"""
		Returns all fields for the given view (also hidden columns, secret columns only if enough privs)
		as array, each field as a tuple (field_name,field_type,default_value,possible_values).
		Caches blobs and fields in session.
		"""
		session = self.REQUEST.SESSION

		if session.has_key('fields') and session['fields'].has_key(view) :
			fields = session['fields'][view]
			if self.userIsAdmin() < 2 :
				fields = [x for x in fields if x[0] not in self.secretColumns()]
		else :
			describe = self.describeTable()
			fields = []
			blobs = []
			for field in describe :
				name = field['Field']
				type = field['Type'].split('(')[0]
				default = field['Default']
				if type == 'enum' or type == 'set' :
					values = field['Type'].split('(')[1]
					values = values[1:-2]
					values = values.split("','")
					for i in range(len(values)) :
						values[i] = values[i].replace("''","'")
						values[i] = values[i].replace("\\\\","\\")
				else :
					values = None
				fields.append( (name,type,default,values) )
				if type in ['tinyblob','blob','mediumblob','longblob'] :
					blobs.append(name)
			# sort fields
			fields.sort(lambda x,y : self.columnSort(x[0],y[0],view))
			if not session.has_key('fields') :
				session['fields'] = {}
			session['fields'][view] = fields
			session['blobs'] = blobs
			if self.userIsAdmin() < 2 :
				fields = [x for x in fields if x[0] not in self.secretColumns()]
		return fields

	security.declareProtected(Permissions.access_contents_information, 'getPrimaryKeys')
	def getPrimaryKeys(self):
		"""
		Returns an array with all primary keys.
		"""
		session = self.REQUEST.SESSION

		if session.has_key('primaryKeys') :
			return session['primaryKeys']
		else :
			fields = self.describeTable()
			keys = [row['Field'] for row in fields if row['Key']=='PRI']
			if not keys :
				keys = [row['Field'] for row in fields]
			session['primaryKeys'] = keys
			return keys

	security.declareProtected(Permissions.access_contents_information, 'getUniqueKeys')
	def getUniqueKeys(self):
		"""
		Returns an array with all unique keys.
		"""
		session = self.REQUEST.SESSION

		if session.has_key('uniqueKeys') :
			return session['uniqueKeys']
		else :
			fields = self.describeTable()
			keys = {}
			for field in fields :
				if field['Key']=='PRI' or field['Key']=='UNI' :
					keys[field['Field']] = field['Extra']
			session['uniqueKeys'] = keys
    		return keys

	security.declareProtected(Permissions.access_contents_information, 'saveDb')
	def saveDb(self, db) :
		"""
		Saves the given database to the session and resets other session data
		"""
		request = self.REQUEST
		session = request.SESSION
		## new db selected -> reset session data
		keys = ['selectedDatabase','selectedTable','primaryKeys','uniqueKeys','selectedColumns','b_start','where','sort','fields','blobs']
		for key in keys :
			if session.has_key(key) :
				del session[key]
		if db :
		# do not save db if empty string
			session['selectedDatabase'] = db

	security.declareProtected(Permissions.access_contents_information, 'saveTable')
	def saveTable(self, table) :
		"""
		Saves the given table to the session and resets other session data
		"""
		request = self.REQUEST
		session = request.SESSION
		## new table selected -> reset session data
		keys = ['selectedTable','primaryKeys','uniqueKeys','selectedColumns','b_start','where','sort','fields','blobs']
		for key in keys :
			if session.has_key(key) :
				del session[key]
		if table :
		# do not save table if empty string
			session['selectedTable'] = table

	security.declareProtected(Permissions.access_contents_information, 'sessionHandling')
	def sessionHandling(self):
		"""
		Does session handling (saves selected db and table, search etc. in session).
		"""
		request = self.REQUEST
		session = request.SESSION
		sessionVariables = getattr(request,'sessionVariables',None)
		loadCookie = getattr(request,'selectedCookie',None)
		saveCookie = getattr(request,'saveCookieName',None)
		cols = getattr(request,'selectedColumns',None)
		db = getattr(request,'selectedDatabase',None)
		table = getattr(request,'selectedTable',None)
		sort = getattr(request,'sort',None)
		b_start = getattr(request,'b_start',None)
		batchSize = getattr(request,'batchSize',None)

		if sessionVariables :
		# load session from permalink
			# delete old keys
			keys = ['selectedDatabase','selectedTable','primaryKeys','uniqueKeys','selectedColumns','b_start','where','sort','fields','blobs']
			for key in keys :
				if session.has_key(key) :
					del session[key]
			# load new keys from permalink
			sessionVariables = marshal.loads(sessionVariables)
			for key in sessionVariables :
				session[key] = sessionVariables[key]

		if loadCookie :
			if loadCookie.startswith('load_') :
			# load session from cookie
				cookie = loadCookie[5:]
				self.loadCookie(cookie)
			elif loadCookie.startswith('delete_') :
			# load session from cookie
				cookie = loadCookie[7:]
				self.deleteCookie(cookie)

		if saveCookie :
		# save session to cookie
			self.saveSessionToCookie(saveCookie)

		if cols :
		# list of selected columns has changed
#			if not same_type(cols,[]) :
			if not isinstance(cols,list) :
				cols = [cols]
			session['selectedColumns'] = {'list': cols}

		if db != None:
		# new db selected ('' is also a new selection!)
			self.saveDb(db)

		if table != None :
		# new table selected ('' is also a new selection!)
			self.saveTable(table)

		if sort and not sessionVariables :
		# new sort specified (do not reverse sort if sessionVariables (permalink) are passed and sort is also in permalink)
		  if session.has_key('sort') and session['sort'][0] == sort :
			if session['sort'][1] == 'ASC' :
			  newSort = (sort,'DESC')
			else :
			  newSort = (sort,'ASC')
		  else :
			newSort = (sort,'ASC')
		  session['sort'] = newSort

		if b_start != None :
		# new batch start selected (0 is also a new batch start!)
		  if b_start :
			session['b_start'] = b_start
		  else :
			if session.has_key('b_start') :
			  del session['b_start']


		if batchSize :
		# new batch size selected
		  if batchSize == 'all' :
			session['batchSize'] = batchSize
		  else :
			session['batchSize'] = int(batchSize)


		###################
		# search requests #
		###################
		if getattr(request,'search',None)=='yes' :
		# advanced search submitted
		## delete batch start
		  if session.has_key('b_start') :
			del session['b_start']
		## get where query
		  fields = [x[0] for x in self.getFields()]
		  where = []
		  for field in fields :
			fieldValue = getattr(request,'fields_'+field, '')
#			if same_type(fieldValue,[]) :
			if isinstance(fieldValue,list) :
			  fieldValue = ','.join(fieldValue)
			compareType = getattr(request,'compareType'+field, None)
			if compareType == 'eq' :
			  operation = '='
			elif compareType == 'ne' :
			  operation = '!='
			elif compareType == 'gt' :
			  operation = '>'
			elif compareType == 'lt' :
			  operation = '<'
			elif compareType == 'ge' :
			  operation = '>='
			elif compareType == 'le' :
			  operation = '<='
			elif compareType == 'cy' :
			  operation = 'LIKE'
			  fieldValue = '%'+fieldValue+'%'
			elif compareType == 'cn' :
			  operation = 'NOT LIKE'
			  fieldValue = '%'+fieldValue+'%'
			if fieldValue and fieldValue != '%%' :
			  where.append({'column':field,'value':fieldValue,'operation':operation})
		## save where in session
		  if where :
			if getattr(request,'subsearch',None) and session.has_key('where') :
			# append search
			  session['where'] = session['where'] + where
			else :
			# redefine search
			  session['where'] = where
		  else :
			if session.has_key('where') and not getattr(request,'subsearch',None) :
			  del session['where']
		elif request.has_key('fieldName') and request.fieldName == 'fullTextSearch' and request.has_key('fullTextSearch') :
		# fulltext search
		## delete batch start
		  if session.has_key('b_start') :
			del session['b_start']
		## get where query
		  compareType = getattr(request,'compareTypefullTextSearch',None)
		  value = getattr(request,'fullTextSearch',None)
		  if compareType == 'cy' :
			where = 'MATCH (`'
		  elif compareType == 'cn' :
			where = 'NOT MATCH (`'
		  where += '`, `'.join([x[0] for x in self.getFields()])
		  where += '`) AGAINST (\''
		  where += value
		  where += '\' IN BOOLEAN MODE)'
		  where = [where]
		## save where in session
		  if getattr(request,'subsearch',None) and session.has_key('where') :
		  # append search
			session['where'] = session['where'] + where
		  else :
		  # redefine search
			session['where'] = where
		elif request.has_key('fieldName') :
		# fast search
		## delete batch start
		  if session.has_key('b_start') :
			del session['b_start']
		## get where query
		  fieldName = getattr(request,'fieldName',None)
		  if not fieldName :
		  # no field selected -> show all rows
			if session.has_key('where') and not getattr(request,'subsearch',None) :
			  del session['where']
		  else :
			fieldValue = getattr(request,'fields_'+fieldName)
			compareType = getattr(request,'compareType'+fieldName,None)
			if compareType == 'eq' :
			  operation = '='
			elif compareType == 'ne' :
			  operation = '!='
			elif compareType == 'gt' :
			  operation = '>'
			elif compareType == 'lt' :
			  operation = '<'
			elif compareType == 'ge' :
			  operation = '>='
			elif compareType == 'le' :
			  operation = '<='
			elif compareType == 'cy' :
			  operation = 'LIKE'
			  fieldValue = '%'+fieldValue+'%'
			elif compareType == 'cn' :
			  operation = 'NOT LIKE'
			  fieldValue = '%'+fieldValue+'%'
			where = [{'column':fieldName,'value':fieldValue,'operation':operation}]
		  ## save where in session
			if getattr(request,'subsearch',None) and session.has_key('where') :
			  session['where'] = session['where'] + where
			else :
			  session['where'] = where
		elif getattr(request,'search',None)=='no' :
		# show all rows
		## delete batch start
		  if session.has_key('b_start') :
			del session['b_start']
		## delete where from session
		  if session.has_key('where') :
			del session['where']

	security.declareProtected(Permissions.access_contents_information, 'getCharacterSetConnection')
	def getCharacterSetConnection(self) :
		"""
		returns debugging information
		"""
		#import sys
		r = ''
		#r += 'default encoding (python): %s<br />' % sys.getdefaultencoding()
		result = self.queryString('show variables like "character_set%"')
		for key,value in result :
			r += '%s: %s<br />\n' % (key,value)
		result = self.queryString('show variables like "collation%"')
		for key,value in result :
			r += '%s: %s<br />\n' % (key,value)

		return r


## add object methods
InitializeClass(DBEditor)

manage_addDBEditorForm = PageTemplateFile(
        'www/DBEditorAdd', globals(),
        __name__ = 'manage_addDBEditorForm')

def manage_addDBEditor(dispatcher, id, title, dbConnectionId='', REQUEST=None):
	"""
	Create a new DBEditor and add it to myself.
	"""
	p = DBEditor(id, unicode(title, 'UTF-8'), unicode(dbConnectionId, 'UTF-8'))
	dispatcher.Destination()._setObject(id, p)
	add_and_edit(dispatcher, id, REQUEST)
	return ''
